#!/usr/bin/python
# -*- coding:Utf-8 -*-

from holygrail import _Realm, _Quest, _Mission, _TagMission
from holygrail import *

VERSION="0.1.2 Galahad"

if __name__ == "__main__":
    pass

# vim:set shiftwidth=4 tabstop=4 expandtab:
